import asyncio
import subprocess
import re
import sys
import urllib.request
from pathlib import Path
from typing import Optional, Dict, Callable, Union
from urllib.parse import quote
from urllib.error import HTTPError
from .data import BrowserInstanceInfo


def make_request(url: str, method="GET") -> str:
    req = urllib.request.Request(url, method=method)
    try:
        with urllib.request.urlopen(req) as response:
            return response.read().decode("utf-8")
    except HTTPError as e:
        return f"{e.code}: {e.reason}"


def find_browser_executable_path(exe="chrome") -> str:
    """ Возвращает путь до EXE. """
    if not sys.platform == "win32":
        raise OSError("find_browser_executable_path() is only available on Windows")

    if exe == "chromium":
        import os
        from pathlib import Path

        app_data = Path(os.getenv('APPDATA')).parent
        browser_path = app_data / "Local/Chromium/Application/chrome.exe"
        if not browser_path.exists():
            return ""
        return browser_path.as_posix()

    import winreg

    reg_path = f"HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\{exe}.exe"
    key, path = re.findall(r"(^[^\\/]+)[\\/](.*)", reg_path)[0]
    connect_to = eval(f"winreg.{key}")
    try: h_key = winreg.OpenKey( winreg.ConnectRegistry(None, connect_to), path )
    except FileNotFoundError: return ""
    result = winreg.QueryValue(h_key, None)
    winreg.CloseKey(h_key)
    return result


def log(data: any = "", lvl: str = "[<- V ->]", eol: str = "\n") -> None:
    print(f"\x1b[32m{lvl} \x1b[38m\x1b[3m{data}\x1b[0m", end=eol)


def save_img_as(path: Union[str, Path], data: bytes) -> None:
    """ Сохраняет набор байт возвращаемый из conn.extend.makeScreenshot(), как изображение.
    :param path:    Путь, или имя файла сохраняемого изображения.
    :param data:    Данные изображения.
    """
    with open(path, "wb") as f:
        f.write(data)


async def async_util_call(function: Callable, *args) -> any:
    """ Позволяет выполнять неблокирующий вызов блокирующих функций. Например:
    await async_util_call(
        save_img_as, "ScreenShot.png", await conn.extend.makeScreenshot()
    )
    """
    return await asyncio.get_running_loop().run_in_executor(
        None, function, *args
    )


def find_instances(for_port: Optional[int] = None, browser: str = "chrome") -> Dict[int, BrowserInstanceInfo]:
    """ !!! ВНИМАНИЕ !!! На Windows 11 может быть отключен компонент WMI("Windows Management
    Instrumentation"), его нужно либо включить в разделе “Программы и компоненты” панели
    управления, либо установить из официальных источников.

    Используется для обнаружения уже запущенных браузеров в режиме отладки.
    Например:
            if browser_instances := find_instances():
                port, instance_info = [(k, v) for k, v in browser_instances.items()][0]
                browser_instance = Browser(instance_info=instance_info)
            else:
                browser_instance = Browser()

            # Или для конкретного, известного порта:
            if browser_instances := find_instances(port):
                instance_info = browser_instances[port]
                browser_instance = Browser(instance_info=instance_info)
            else:
                browser_instance = Browser()
    :param for_port:    - порт, для которого осуществляется поиск.
    :param browser:     - браузер, для которого запрашивается поиск.
    :return:
    """
    win_exp = re.compile(r"^\"[^\"]+(?<=\\)(\w+\.\w+)\".*?--remote-debugging-port=(\d+).*?(\d+)\s*$")
    linux_exp = re.compile(r"^[/\w]+/(\w+).*?--remote-debugging-port=(\d+)")

    result = {}
    if sys.platform == "win32":
        if "chrom" in browser: browser = "chrome.exe"
        elif "brave" in browser: browser = "brave.exe"
        elif "edge" in browser: browser = "msedge.exe"
        # else: ValueError("Not support browser: " + browser)
        cmd = f"WMIC PROCESS WHERE NAME='{browser}' GET Commandline,Processid"
        for line in subprocess.Popen(cmd, stdout=subprocess.PIPE).stdout:
            if b"--type=renderer" not in line and b"--remote-debugging-port=" in line:
                name, port, pid = win_exp.findall(line.decode())[0]
                inst = BrowserInstanceInfo(
                    name=name,
                    port=(port := int(port)),
                    pid=int(pid),
                    headless=b"--headless" in line
                )
                if for_port == port:
                    return {port: inst}
                result[port] = inst

    elif sys.platform == "linux":
        if "chrom" in browser: browser = "chrome"
        elif "brave" in browser: browser = "brave"
        elif "edge" in browser: browser = "msedge"
        # else: ValueError("Not support browser: " + browser)
        try: itr = map(int, subprocess.check_output(["pidof", browser]).split())
        except subprocess.CalledProcessError: itr = []
        for pid in itr:
            with open("/proc/" + str(pid) + "/cmdline") as f: cmd_line =  f.read()[:-1]
            if "--type=renderer" not in cmd_line and "--remote-debugging-port=" in cmd_line:
                name, port = linux_exp.findall(cmd_line)[0]
                inst = BrowserInstanceInfo(
                    name=name,
                    port=(port := int(port)),
                    pid=pid,
                    headless="--headless" in cmd_line
                )
                if for_port == port:
                    return {port: inst}
                result[port] = inst
    else:
        raise OSError(f"Platform '{sys.platform}' — not supported")
    return {} if for_port else result


def prepare_url(url: Union[str, bytes, None], browser_name: str, app: bool = False) -> Optional[str]:
    """ Подготавливает строку для передачи в navigate(), или в конструкторе Browser."""

    # Если url == None
    if url is None:
        return "--app=data:text/html," if app else None

    # Если передали строку
    if type(url) is str:

        # И эта строка - пустая
        if url == "":
            return "--app=data:text/html," if app else "about:blank"

        # Но если строка не содержит признаков url-адреса, то
        # считаем это HTML-разметкой
        if not url.startswith(("http", browser_name)):
            url = "data:text/html," + quote(url)
            return "--app=" + url if app else url

        # Иначе - это адрес
        return "--app=" + url if app else url

    # Иначе считаем, что это не конвертированный набор байт
    # из base64, в котором HTML-разметка
    url = "data:text/html;Base64," + url.decode()
    return "--app=" + url if app else url



import asyncio
from typing import Optional, Union, Callable, Awaitable, TYPE_CHECKING
from ...data import DomainEvent
from ...utils import prepare_url
from .types import FrameTree, LifecycleEventData
if TYPE_CHECKING:
    from ...connection import Connection


class Page:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Page
    """
    __slots__ = (
        "_connection", "enabled", "loading_state_watcher_enabled", "network_idle_state_watcher_enabled",
        "lifecycle_events_enabled", "loading_state", "network_idle_state"
    )
    def __init__(self, conn) -> None:
        self._connection: Connection = conn
        self.enabled = False
        self.loading_state_watcher_enabled = False
        self.network_idle_state_watcher_enabled = False
        self.loading_state = asyncio.Event()
        self.lifecycle_events_enabled = False
        self.network_idle_state = asyncio.Event()

    async def enable(self) -> None:
        """
        Включает уведомления домена 'Page'.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-enable
        """
        if not self.enabled:
            await self._connection.call("Page.enable")
            self.enabled = True

    async def disable(self) -> None:
        """
        Выключает уведомления домена 'Page'.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-disable
        """
        if self.enabled:
            if self.loading_state_watcher_enabled:
                await self.enableLoadWatcher(False)
            if self.network_idle_state_watcher_enabled:
                await self.enableNetworkIdleWatcher(False)
            await self._connection.call("Page.disable")
            self.enabled = False

    async def enableLoadWatcher(self, state: bool) -> None:
        """ Уведомляет событие loading_state, что основной фрейм страницы завершил загрузку.
        :param state:           Вкл/выкл
        """
        async def load_watcher_wrapper(params: dict, *_) -> None:
            if params["frameId"] == self._connection.conn_id:
                self.loading_state.set()

        if state != self.loading_state_watcher_enabled:
            if state:
                await self._connection.addListenerForEvent(
                    PageEvent.frameStoppedLoading, load_watcher_wrapper)
            else:
                self._connection.removeListenerForEvent(
                    PageEvent.frameStoppedLoading, load_watcher_wrapper)
            self.loading_state_watcher_enabled = state

    async def createIsolatedWorld(
            self, frameId: str, worldName: Optional[str] = None, grantUniversalAccess: Optional[bool] = None) -> int:
        """
        Создаёт изолированный мир для указанного фрейма.
        https://chromedevtools.github.io/devtools-protocol/1-3/Page/#method-createIsolatedWorld
        :param frameId:                 Идентификатор фрейма, в котором должен быть создан изолированный мир.
        :param worldName:               Необязательное имя, о котором сообщается в контексте выполнения.
        :param grantUniversalAccess:    Должен ли быть предоставлен универсальный доступ к изолированному миру.
                                            Это мощная опция, используйте ее с осторожностью.
        :return:            Runtime.ExecutionContextId
        """
        args = dict(frameId=frameId)
        if worldName is not None: args.update(worldName=worldName)
        if grantUniversalAccess is not None: args.update(grantUniversalAccess=grantUniversalAccess)
        result: dict = await self._connection.call("Page.createIsolatedWorld", args)
        return result.get("executionContextId")

    async def getFrameTree(self) -> 'FrameTree':
        """
        Возвращает структуру присутствующих на странице фреймов. !!! Справедливо только в рамках одного домена.
        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-getFrameTree
        """
        result = await self._connection.call("Page.getFrameTree")
        return FrameTree(**result.get("frameTree"))


    async def handleJavaScriptDialog(self, accept: bool, promptText: str = "") -> None:
        """
        Подтверждает или закрывает диалоговое окно, инициированное JavaScript (alert, confirm,
            prompt или для onbeforeunload).
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-handleJavaScriptDialog
        :param accept:          'True' — принять, 'False' — отклонить диалог.
        :param promptText:      (optional) Текст для ввода в диалоговом окне, прежде чем принять.
                                    Используется, только если это диалоговое окно с подсказкой.
        :return:
        """
        args = {"accept": accept}
        if promptText: args.update({"promptText": promptText})
        await self._connection.call("Page.handleJavaScriptDialog", args)

    async def navigate(
            self,
            url:  Union[str,     bytes] = "about:blank",
            wait_for_load:         bool = True,
            wait_for_network_idle: bool = False
    ) -> None:
        """
        Переходит на адрес указанного 'url'.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-navigate
        :param url:                     Адрес, по которому происходит навигация.
        :param wait_for_load:           (optional) Если 'True' - ожидает состояния остановки
                                            загрузки ресурсов, если активны уведомления домена Page.
        :param wait_for_network_idle:   (optional) Если 'True' - ожидает прекращения активности сети
        :return:
        """
        url = prepare_url(url, self._connection.browser_name)

        await self._connection.call("Page.navigate", {"url": url})
        if wait_for_load or wait_for_network_idle:
            await self.waitForLoad(wait_for_load, wait_for_network_idle)

    async def waitForLoad(self, by_load_state: bool = False, by_network_idle: bool = True) -> None:
        """ Дожидается указанного состояния.
        :param by_load_state:       Ожидать завершения загрузки страницы.
        :param by_network_idle:     Ожидать завершения активности сети.
        """

        if not by_load_state and not by_network_idle:
            raise ValueError("'by_load_state' or 'by_network_idle' must be a True")

        if not self.enabled:
            await self.enable()

        if by_load_state:
            if not self.loading_state_watcher_enabled:
                await self.enableLoadWatcher(True)
                self.loading_state_watcher_enabled = True
            await self.loading_state.wait()

        if by_network_idle:
            if not self.network_idle_state_watcher_enabled:
                await self.enableNetworkIdleWatcher(True)
            await self.network_idle_state.wait()

    async def addScriptOnLoad(self, src: str) -> str:
        """
        Запускает полученный 'src' скрипта в каждом фрейме и перед загрузкой скриптов самого фрейма.
            Так же, такой скрипт будет запускаться автоматически в течении всей жизни инстанса,
            включая перезагрузку страницы.
        Требует включения( Enable() ) уведомлений домена "Page".
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-addScriptToEvaluateOnNewDocument
        :param src:             Текст сценария.
        :return:                identifier -> Уникальный идентификатор скрипта.
        """
        if not self.enabled: await self.enable()
        return (await self._connection.call("Page.addScriptToEvaluateOnNewDocument", {"source": src}))["identifier"]

    async def removeScriptOnLoad(self, identifier: str) -> None:
        """
        Удаляет данный скрипт из списка запускаемых при загрузке фрейма.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-removeScriptToEvaluateOnNewDocument
        :param identifier:      Идентификатор сценария.
        :return:
        """
        await self._connection.call("Page.removeScriptToEvaluateOnNewDocument", {"identifier": identifier})

    async def setDocumentContent(self, html: str, frameId: str = None) -> None:
        """
        Устанавливает полученную разметку как HTML-код документа.
            frameId — можно найти среди параметров события 'Runtime.executionContextCreated',
            а так же у корневого элемента документа root.children[1].frameId
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setDocumentContent
        :param html:            HTML-разметка.
        :param frameId:         Идентификатор фрейма, которому предназначается html.
        :return:
        """
        if frameId is None: frameId = self._connection.conn_id
        await self._connection.call("Page.setDocumentContent", {"frameId": frameId, "html": html})

    async def stopLoading(self) -> None:
        """
        Останавливает загрузку страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-stopLoading
        :return:
        """
        await self._connection.call("Page.stopLoading")

    async def setAdBlockingEnabled(self, enabled: bool) -> None:
        """
        (EXPERIMENTAL)
        Включите экспериментальный рекламный фильтр Chrome на всех сайтах.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setAdBlockingEnabled
        :param enabled:         Включить?
        :return:
        """
        if not self.enabled: await self.enable()
        await self._connection.call("Page.setAdBlockingEnabled", {"enabled": enabled})

    async def setFontFamilies(self, fontFamilies: dict) -> None:
        """
        (EXPERIMENTAL)
        Установить общие семейства шрифтов.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setFontFamilies
        :param fontFamilies:    Семейства шрифтов:
                                    {
                                        "standard":   str(), -> (optional)
                                        "fixed":      str(), -> (optional)
                                        "serif":      str(), -> (optional)
                                        "sansSerif":  str(), -> (optional)
                                        "cursive":    str(), -> (optional)
                                        "fantasy":    str(), -> (optional)
                                        "pictograph": str(), -> (optional)
                                    }
        :return:
        """
        await self._connection.call("Page.setFontFamilies", {"fontFamilies": fontFamilies})

    async def setFontSizes(self, fontSizes: dict) -> None:
        """
        (EXPERIMENTAL)
        Установите размеры шрифта по умолчанию.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setFontSizes
        :param fontSizes:       Определяет размеры шрифта для установки. Если размер шрифта не
                                    указан, он не будет изменен:
                                    {
                                        "standard": int(), -> (optional)
                                        "fixed":    int(), -> (optional)
                                    }
        :return:
        """
        await self._connection.call("Page.setFontSizes", {"fontSizes": fontSizes})

    async def captureScreenshot(
        self,
        format_: str = "",
        quality: int = -1,
        clip: Optional[dict] = None,
        fromSurface: bool = True
    ) -> str:
        """
        Сделать скриншот. Возвращает кодированное base64 представление скриншота.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-captureScreenshot
        :param format_:         string => Image compression format (defaults to png) (jpeg or png).
        :param quality:         integer => Compression quality from range [0..100] (jpeg only).
        :param clip:            {
                                    "x": "number => X offset in device independent pixels (dip).",
                                    "y": "number => Y offset in device independent pixels (dip).",
                                    "width": "number => Rectangle width in device independent pixels (dip).",
                                    "height": "number => Rectangle height in device independent pixels (dip).",
                                    "scale": "number => Page scale factor."
                                }
        :param fromSurface:     boolean => Capture the screenshot from the surface, rather than the view.
                                    Defaults to true.
        :return:                string => Base64-encoded image data.
        """
        args = {"fromSurface": fromSurface}
        if format_: args.update({"format": format_})
        if quality > -1 and format_ == "jpeg": args.update({"quality": quality})
        if clip: args.update({"clip": clip})
        return (await self._connection.call("Page.captureScreenshot", args))["data"]

    async def printToPDF(
        self,
        landscape:               Optional[bool] = None,
        displayHeaderFooter:     Optional[bool] = None,
        printBackground:         Optional[bool] = None,
        scale:                  Optional[float] = None,
        paperWidth:             Optional[float] = None,
        paperHeight:            Optional[float] = None,
        marginTop:              Optional[float] = None,
        marginBottom:           Optional[float] = None,
        marginLeft:             Optional[float] = None,
        marginRight:            Optional[float] = None,
        pageRanges:               Optional[str] = None,
        ignoreInvalidPageRanges: Optional[bool] = None,
        headerTemplate:           Optional[str] = None,
        footerTemplate:           Optional[str] = None,
        preferCSSPageSize:       Optional[bool] = None,
        transferMode:             Optional[str] = None
    ) -> dict:
        """
        Печатает страницу как PDF.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-printToPDF
        :param landscape:               (optional) Ориентация бумаги. По умолчанию false.
        :param displayHeaderFooter:     (optional) Отобразить header и footer. По умолчанию false.
        :param printBackground:         (optional) Печать фоновой графики. По умолчанию false.
        :param scale:                   (optional) Масштаб рендеринга веб-страницы. По умолчанию 1.
        :param paperWidth:              (optional) Ширина бумаги в дюймах. По умолчанию 8,5 дюймов.
        :param paperHeight:             (optional) Высота бумаги в дюймах. По умолчанию 11 дюймов.
        :param marginTop:               (optional) Верхний отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param marginBottom:            (optional) Нижний отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param marginLeft:              (optional) Левый отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param marginRight:             (optional) Правый отступ в дюймах. По умолчанию 1 см (~ 0,4 дюйма).
        :param pageRanges:              (optional) Диапазон бумаги для печати, например, '1-5, 8, 11-13'. По
                                            умолчанию используется пустая строка, что означает печать всех страниц.
        :param ignoreInvalidPageRanges: (optional) Следует ли игнорировать недействительные, но успешно
                                            проанализированные диапазоны страниц, например '3-2'. По умолчанию false.
        :param headerTemplate:          (optional) HTML-шаблон для печатного заголовка. Должна быть допустимая
                                            разметка HTML со следующими классами, используемыми для вставки
                                            значений печати в них:
                                                * date:       formatted print date
                                                * title:      document title
                                                * url:        document location
                                                * pageNumber: current page number
                                                * totalPages: total pages in the document
                                                    Например, <span class=title></span> будет генерировать span,
                                                        содержащий заголовок.
        :param footerTemplate:          (optional) HTML-шаблон для печати 'footer'. Следует использовать тот же
                                            формат, что и headerTemplate.
        :param preferCSSPageSize:       (optional) Предпочитать или нет размер страницы, как определено CSS. По
                                            умолчанию установлено значение false, и в этом случае содержимое будет
                                            масштабироваться по размеру бумаги.
        :param transferMode:            (optional, EXPERIMENTAL) вернуть как поток. Допустимые значения:
                                            ReturnAsBase64, ReturnAsStream
        :return:                        {
                                            "data": str(Данные PDF, кодированные в Base64.),
                                                ->  Будет пустым если в 'transferMode' выбрано 'ReturnAsStream'.
                                            "stream": str(Это либо получается из другого метода, либо указывается как blob <uuid> это UUID Blob.)
                                                -> StreamHandle
                                        }
        """
        args = {}
        if landscape is not None:               args.update({"landscape": landscape})
        if displayHeaderFooter is not None:     args.update({"displayHeaderFooter": displayHeaderFooter})
        if printBackground is not None:         args.update({"printBackground": printBackground})
        if scale is not None:                   args.update({"scale": scale})
        if paperWidth is not None:              args.update({"paperWidth": paperWidth})
        if paperHeight is not None:             args.update({"paperHeight": paperHeight})
        if marginTop is not None:               args.update({"marginTop": marginTop})
        if marginBottom is not None:            args.update({"marginBottom": marginBottom})
        if marginLeft is not None:              args.update({"marginLeft": marginLeft})
        if marginRight is not None:             args.update({"marginRight": marginRight})
        if pageRanges is not None:              args.update({"pageRanges": pageRanges})
        if ignoreInvalidPageRanges is not None: args.update({"ignoreInvalidPageRanges": ignoreInvalidPageRanges})
        if headerTemplate is not None:          args.update({"headerTemplate": headerTemplate})
        if footerTemplate is not None:          args.update({"footerTemplate": footerTemplate})
        if preferCSSPageSize is not None:       args.update({"preferCSSPageSize": preferCSSPageSize})
        if transferMode is not None:            args.update({"transferMode": transferMode})

        return await self._connection.call("Page.printToPDF", args)

    async def reload(
            self,
            ignoreCache: bool = False,
            scriptToEvaluateOnLoad: str = "",
            wait_for_load: bool = False,
            wait_for_network_idle: bool = True
    ) -> None:
        """
        Перезагружает страницу инстанса, при необходимости игнорируя кеш.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-reload
        :param ignoreCache:             (optional) Игнорировать кеш?
        :param scriptToEvaluateOnLoad:  (optional) Если установлено, сценарий будет вставлен во все
                                            фреймы проверяемой страницы после перезагрузки. Аргумент
                                            будет игнорироваться при перезагрузке источника dataURL.
        :param wait_for_load:           (optional) По умолчанию — дожидается полного завершения
                                            загрузки страницы(document.readyState === "complete").
                                            Установите False, если это поведение не требуется.
        :return:
        """
        if self.enabled:
            self.loading_state.clear()
        if self.lifecycle_events_enabled:
            self.network_idle_state.clear()
        args = {}
        if ignoreCache:            args.update({"ignoreCache": ignoreCache})
        if scriptToEvaluateOnLoad: args.update({"scriptToEvaluateOnLoad": scriptToEvaluateOnLoad})
        await self._connection.call("Page.reload", args)
        if wait_for_load or wait_for_network_idle:
            await self.waitForLoad(wait_for_load, wait_for_network_idle)

    async def getNavigationHistory(self) -> dict:
        """
        Возвращает историю навигации для текущей страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-getNavigationHistory
        :return:            {
                                "currentIndex": int(), -> Индекс текущей записи истории навигации.
                                "entries": list({
                                    "id": int(),  -> Уникальный идентификатор записи истории навигации.
                                    "url": str(), -> URL записи истории навигации.
                                    "userTypedURL": str(), -> URL, который пользователь ввел в строке URL.
                                    "title": str(), -> Название записи истории навигации.
                                    "transitionType": str(
                                        link, typed, address_bar, auto_bookmark, auto_subframe,
                                        manual_subframe, generated, auto_toplevel, form_submit,
                                        reload, keyword, keyword_generated, other
                                    ) -> Тип перехода.
                                }, ... ) -> Список записей истории навигации.
                            }
        """
        return await self._connection.call("Page.getNavigationHistory")

    async def navigateToHistoryEntry(self, entryId: int) -> None:
        """
        Навигация текущей страницы к выбранной записи в истории навгации.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-navigateToHistoryEntry
        :param entryId:             Уникальный идентификатор записи для перехода.
        :return:
        """
        if self.enabled: self.loading_state.clear()
        await self._connection.call("Page.navigateToHistoryEntry", {"entryId": entryId})

    async def resetNavigationHistory(self) -> None:
        """
        Сбрасывает историю навигации для текущей страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-resetNavigationHistory
        :return:
        """
        await self._connection.call("Page.resetNavigationHistory")

    async def setInterceptFileChooserDialog(self, enabled: bool) -> None:
        """
        Перехватывает запросы выбора файлов и передает управление клиентам протокола. Когда включен перехват
            файлов, диалоговое окно выбора файлов не отображается. Вместо этого генерируется событие
            протокола Page.fileChooserOpened.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setInterceptFileChooserDialog
        :param enabled:             Включить перехват?
        :return:
        """
        await self._connection.call("Page.setInterceptFileChooserDialog", {"enabled": enabled})

    async def setLifecycleEventsEnabled(
            self, enabled: bool,
            handler: Optional[Callable[[LifecycleEventData], Awaitable[None]]] = None,
    ) -> None:
        """
        Определяет, будет ли страница генерировать события жизненного цикла.
        https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-setLifecycleEventsEnabled
        :param enabled:            Если true, начинает генерировать события жизненного цикла.
        :param handler:            Awaitable объект, которому будут переданы данные события.
        :return:
        """
        async def life_cycle_event_wrapper(params: dict, *_) -> None:
            await handler(LifecycleEventData(**params))

        if self.lifecycle_events_enabled != enabled:
            if enabled and handler is not None:
                await self._connection.addListenerForEvent(
                    PageEvent.lifecycleEvent, life_cycle_event_wrapper)
            else:
                self._connection.removeListenersForEvent(PageEvent.lifecycleEvent)

            await self._connection.call("Page.setLifecycleEventsEnabled", {"enabled": enabled})
            self.lifecycle_events_enabled = enabled

    async def enableNetworkIdleWatcher(self, state: bool) -> None:
        """ [Вк/Вык]лючает наблюдение за событием жизненного цикла ресурса, когда
        у него возникает состояние с именем `networkIdle`, говорящее о простое сети.

        Этот метод активируется автоматически при совершении переходов по URL-адресам
        """
        async def idle_watcher_wrapper(params: dict, *_) -> None:
            if params["frameId"] == self._connection.conn_id:
                if params["name"] == "networkIdle":
                    self.network_idle_state.set()

        if state:
            if not self.lifecycle_events_enabled:
                await self.setLifecycleEventsEnabled(True)
            if not self.network_idle_state_watcher_enabled:
                await self._connection.addListenerForEvent(
                    PageEvent.lifecycleEvent, idle_watcher_wrapper)
        elif not state and self.network_idle_state_watcher_enabled:
            self._connection.removeListenerForEvent(
                PageEvent.lifecycleEvent, idle_watcher_wrapper)
        self.network_idle_state_watcher_enabled = state

    async def bringToFront(self) -> None:
        """
        Выводит страницу на передний план (активирует вкладку).
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-bringToFront
        :return:
        """
        await self._connection.call("Page.bringToFront")


class PageEvent(DomainEvent):
    domContentEventFired = "Page.domContentEventFired"
    fileChooserOpened = "Page.fileChooserOpened"
    frameAttached = "Page.frameAttached"
    frameDetached = "Page.frameDetached"
    frameNavigated = "Page.frameNavigated"
    interstitialHidden = "Page.interstitialHidden"
    interstitialShown = "Page.interstitialShown"
    javascriptDialogClosed = "Page.javascriptDialogClosed"
    javascriptDialogOpening = "Page.javascriptDialogOpening"
    lifecycleEvent = "Page.lifecycleEvent"
    loadEventFired = "Page.loadEventFired"
    windowOpen = "Page.windowOpen"
    frameClearedScheduledNavigation = "Page.frameClearedScheduledNavigation"    # DEPRECATED
    frameScheduledNavigation = "Page.frameScheduledNavigation"                  # DEPRECATED
    backForwardCacheNotUsed = "Page.backForwardCacheNotUsed"                    # ! EXPERIMENTAL
    compilationCacheProduced = "Page.compilationCacheProduced"                  # ! EXPERIMENTAL
    documentOpened = "Page.documentOpened"                                      # ! EXPERIMENTAL
    frameRequestedNavigation = "Page.frameRequestedNavigation"                  # ! EXPERIMENTAL
    frameResized = "Page.frameResized"                                          # ! EXPERIMENTAL
    frameStartedLoading = "Page.frameStartedLoading"                            # ! EXPERIMENTAL
    frameStoppedLoading = "Page.frameStoppedLoading"                            # ! EXPERIMENTAL
    navigatedWithinDocument = "Page.navigatedWithinDocument"                    # ! EXPERIMENTAL
    screencastFrame = "Page.screencastFrame"                                    # ! EXPERIMENTAL
    screencastVisibilityChanged = "Page.screencastVisibilityChanged"            # ! EXPERIMENTAL
    downloadProgress = "Page.downloadProgress"                                  # * EXPERIMENTAL DEPRECATED
    downloadWillBegin = "Page.downloadWillBegin"                                # * EXPERIMENTAL DEPRECATED



"""
    Набор готового JS-функционала.
"""


# Конвертирует секунды в строку вида:
#   parse_sec(1) => 00:00:01
# Игнорирует дни при переполнении ёмкости:
#   parse_sec(86_401) => 00:00:01
PARSE_SEC = """
function parse_sec(sec) {
    return new Date(Math.abs(sec) * 1000).toISOString().substr(11, 8);
}"""

# Конвертирует секунды в строку. Отображает знак:
#   parse_sec(-1) => -00:00:01
# Игнорирует дни при переполнении ёмкости:
#   parse_sec(-86_401) => -00:00:01
PARSE_SEC_SIGN = """
function parse_sec_sign(sec) {
    return (sec < 0 ? "-" : "") + new Date(Math.abs(sec) * 1000).toISOString().substr(11, 8);
}"""

# Отображает дни при переполнении ёмкости, или days === true:
#   parse_sec_days(86_401) => 01д 00:00:01
#   parse_sec_days(6_401) => 01:46:41
PARSE_SEC_DAYS = """
function parse_sec_days(sec, days=false) {
    const d = (new Date(Math.abs(sec) * 1000) - new Date(0)) /864e5|0;
    const p = new Date(Math.abs(sec) * 1000).toISOString().substr(11, 8);
    return (d || days ? String(d).padStart(2,"0") + "д " : "") + p;
}"""

# Отображает дни при переполнении ёмкости, или days === true. Отображает знак:
#   parse_sec_days(-86_401) => -01д 00:00:01
#   parse_sec_days(-6_401) => -01:46:41
PARSE_SEC_DAYS_SIGN = """
function parse_sec_days_sign(sec, days=false) {
    const d = (new Date(Math.abs(sec) * 1000) - new Date(0)) /864e5|0;
    const p = new Date(Math.abs(sec) * 1000).toISOString().substr(11, 8);
    return (sec < 0 ? "-" : "") + (d || days ? String(d).padStart(2,"0") + "д " : "") + p;
}"""


# Создаёт элемент <style> и заполняет его правилами, пришедшими в base64-строке
ADD_RULES_FROM_B64 = """
function add_rules_from_b64(b64_rules) {
    const style = document.createElement("style");
    style.appendChild(document.createTextNode(atob(b64_rules)));
    document.head.appendChild(style);
}"""

# Создаёт элемент <style> и заполняет его правилами, пришедшими в строке
ADD_RULES_FROM_STR = """
function add_rules(str_rules) {
    const style = document.createElement("style");
    style.appendChild(document.createTextNode(str_rules));
    document.head.appendChild(style);
}"""

# Получает самый последний styleSheet и вставляет в него правило самым последним
INJECT_RULE = """
function inject_rule(rule) {
    const sheets = document.styleSheets;
    const last_sheet = sheets.length - 1;
    const to_last_pos = sheets[last_sheet].length;
    sheets[last_sheet].insertRule(rule, to_last_pos);
}"""

# Возвращает полный CSS-путь(селектор) до переданного элемента
#   от ближайшего идентификатора, или body, если не будет встречен по пути
#   #uniq-id>:nth-child(3)>:nth-child(1)>:nth-child(1)>:nth-child(27)> ...
CSS_PATH = """
function css_path(e) {
    if(e.id)return "#"+e.id;if(e.tagName.toLowerCase()==="body")return "body";
    let n=1,s=e;while(s&&(s=s.previousElementSibling)){n+=1;}
    return cssPath(e.parentNode)+">:nth-child("+n+")";
}"""

# Создаёт textarea элемент и размещает его в body за краем вьюпорта, а
# так же объявляет функцию clip_copy() отправляющую текст в буфер обмена
CLIP_COPY = """
const CLIP_NODE = document.createElement("textarea");
CLIP_NODE.value = "";
Object.assign(CLIP_NODE.style, { position: "fixed", left: "-1000px", top: "-1000px" });
document.body.appendChild(CLIP_NODE);

function clip_copy(text) {
    CLIP_NODE.value = text;
    CLIP_NODE.select();
    document.execCommand("copy");
    CLIP_NODE.value = ""
}"""

# Динамически создаёт текстовый файл, наполняет его контентом
#   и предлагает скачать, вызывая соответствующий диалог.
DYNAMIC_CREATE_TXT_FILE = """
function dyn_file_create(content, name="file.txt") {
    const link = document.createElement("a");
    link.download = name;
    const blob = new Blob([content], {type: "text/plain;charset=utf-8;"});
    link.href = URL.createObjectURL(blob);
    link.click();
    URL.revokeObjectURL(link.href);
}"""

# Возвращает Base64 строку для переданного <img ...>
IMG_TO_B64 = """
function image_to_b64(img_node) {
    const canvas = document.createElement("canvas");
    canvas.height = img_node.naturalHeight;
    canvas.width = img_node.naturalWidth;
    canvas.getContext("2d").drawImage(img_node, 0, 0, canvas.width, canvas.height);
    return canvas.toDataURL().replace("data:image/png;base64,", "");
}"""

# Возвращает простой хэш строки
HASH_CODE = """
String.prototype.hashCode=function(){return this.split("").reduce((a,b)=>{
a=((a<<5)-a)+b.charCodeAt(0);return a&a},0);}"""

# Добавляет строкам прототип format
# "{0} is dead, but {1} is alive! {0} {2}".format(111, "ASP.NET", "tri", "chetyre")
STRING_FORMAT = r"""
String.prototype.format=function(){const a = arguments;
return this.replace(/{(\d+)}/g,function(m,n){return typeof a[n]!='undefined'?a[n]:m;});};"""

# Первую букву слова заглавной
#   "string data".capitalize() => "String data"
CAPITALIZE = r"""
String.prototype.capitalize=function(){return this.replace(/^\w/,function(c){return c.toUpperCase()})}"""

# Перемешивание для массивов
ARR_SHUFFLE = """
Array.prototype.shuffle=function(){for (let i=this.length-1;i>0;i--){
let t,j=Math.floor(Math.random()*(i+1));t=this[i];this[i]=this[j];this[j]=t;}}"""

# Возвращает Y-координату элемента на странице, или текущее
#   положение вертикального скролла, если элемент не указан.
GET_Y = """
function get_y(node=null){return pageYOffset+(node!==null ? node.getBoundingClientRect().y:0);}"""

# random
RANDINT = """
function randint(min,max){return Math.floor(min+Math.random()*(max+1-min))}"""

UNIFORM = """
function uniform(min,max){return Math.random()*(max-min)+min}"""

# function get_roots(el) - возвращает список всех shadowDOM-потомков
#   элемента, не включая их потомков.
# function in_root(root_element, selector) - возвращает найденный по
#   селектору элемент у root-элемента.
# Рекурсия in_root(), не погружается настолько, чтобы найти очень
#   глубокие элементы, поэтому её приходится запускать вручную несколько раз:
#       const settings = in_root(get_roots($("body"))[0], "settings-privacy-page");
#       const category = in_root(settings, "category-default-setting");
#       const control  = in_root(category, "#control");
#       control.click();
SHADOW_DOM = """
function get_roots(el) {
    if (!!el.shadowRoot) {
        return [el.shadowRoot];
    } const roots = [];
    [...el.children].forEach((e) => {
        roots.push(...get_roots(e));
    }); return roots;
}

function in_root(root_element, selector) {
    const el = root_element.querySelector(selector);
    if (el) { return el; }
    let find;
    for (let r of get_roots(root_element)) {
        if ((find = in_root(r, selector))) { return find; }
    }
}"""

# Возвращает булево значение описывающее, достигла ли прокрутка страницы
#   максимума высоты прокрутки, или лимита, если его значение больше нуля.
#   true - если корневой элемент === null.
IS_BOTTOM_SCROLL = """
function is_bottom_scroll(limit=10_000) {
    if (!document.documentElement) { return true; }
    const sh = window.pageYOffset + document.documentElement.clientHeight;
    if (limit <= 0) {
        return sh >= document.documentElement.scrollHeight;
    } return sh >= document.documentElement.scrollHeight || sh > limit;
}"""

# Возвращает (JSON) список из двух величин, описывающих текущую прокрутку по
#   X и Y координатам для элемента. Если элемент не передан, вычисляется
#   прокрутка вьюпорта.
HOW_SCROLL = """
function how_scroll(element=null) {
    const root = element || document.documentElement;
    const scrollLeft = window.pageXOffset || document.documentElement.scrollLeft;
    const scrollTop = window.pageYOffset || document.documentElement.scrollTop;
    return JSON.stringify([root.scrollTop, root.scrollLeft]);
}"""

# Асинхронный слип
SLEEP = """
async function sleep(delay=1000) {await new Promise(r => setTimeout(r, delay));}"""

# Перетаскивает элемент относительно текущей позиции
DRAG_NODE = SLEEP + """
async function drag_node(node, x=0, y=0) {
    const options = {
        clientX: node.getBoundingClientRect().left,
        clientY: node.getBoundingClientRect().top,
        bubbles: true, cancelable: true
    };
    node.dispatchEvent(new MouseEvent("mousedown", options));
    await sleep(100);
    options.clientX += x; options.clientY += y;
    node.dispatchEvent(new MouseEvent("mousemove", options));
    await sleep(100);
    node.dispatchEvent(new MouseEvent("mouseup", options));
}"""


def noisify_canvas(*, r: int = 0, g: int = 0, b: int = 0, a: int = 0) -> str:
    """ Возвращает самовызываемый объект, при выполнении которого на странице,
    добавляется шум в канвас.
    https://browserleaks.com/canvas
    """
    canvas_noisify_src = """
    ((r=0, g=0, b=0, a=0) => {{
        const toBlob = HTMLCanvasElement.prototype.toBlob;
        const toDataURL = HTMLCanvasElement.prototype.toDataURL;
        const getImageData = CanvasRenderingContext2D.prototype.getImageData;
        const noisify = function (canvas, context) {{
            const shift = {{'r': r, 'g': g, 'b': b, 'a': a}};
            const width = canvas.width, height = canvas.height;
            const imageData = getImageData.apply(context, [0, 0, width, height]);
            for (let i = 0; i < height; i++) {{
                for (let j = 0; j < width; j++) {{
                    const n = ((i * (width * 4)) + (j * 4));
                    imageData.data[n + 0] = imageData.data[n + 0] + shift.r;
                    imageData.data[n + 1] = imageData.data[n + 1] + shift.g;
                    imageData.data[n + 2] = imageData.data[n + 2] + shift.b;
                    imageData.data[n + 3] = imageData.data[n + 3] + shift.a;
                }}
            }}
            context.putImageData(imageData, 0, 0);
        }}; Object.defineProperty(HTMLCanvasElement.prototype, "toBlob", {{
            "value": function () {{
                noisify(this, this.getContext("2d"));
                return toBlob.apply(this, arguments);
            }}
        }}); Object.defineProperty(HTMLCanvasElement.prototype, "toDataURL", {{
            "value": function () {{
                noisify(this, this.getContext("2d"));
                return toDataURL.apply(this, arguments);
            }}
        }});  Object.defineProperty(CanvasRenderingContext2D.prototype, "getImageData", {{
            "value": function () {{
                noisify(this.canvas, this);
                return getImageData.apply(this, arguments);
            }}
        }});
    }})({r}, {g}, {b}, {a});
    """
    return canvas_noisify_src.format(r=r, g=g, b=b, a=a)
